import string
from typing import Any, Callable, Dict, Optional, Set, Tuple, Type, Union  # noqa

from enum import Enum  # via enum34 package on python 2.7

from popget.errors import ArgNameConflict
from popget.extratypes import ResponseTypes  # noqa


class BodyType(Enum):
    """
    `body_type` of requests which send a body
    """
    JSON = 'json'
    FORM_ENCODED = 'data'


def _validate_arg_name(arg, arg_type, reserved):
    # type: (str, str, Set[str]) -> str
    """
    Validate that `arg` does not conflict with names already defined in
    `reserved`.

    Kwargs:
        arg (str): the arg name to validate
        arg_type (str): a description for the arg type, to use in exception
            message if conflicting
        reserved (Set[str]): reserved arg names

    Returns:
        str: validated arg name
    """
    if arg in reserved:
        raise ArgNameConflict(
            '{arg_type} arg `{arg}` conflicts with a reserved arg name.'.format(
                arg_type=arg_type,
                arg=arg,
            )
        )
    return arg


NO_DEFAULT = object()


class Arg(object):
    """
    Querystring argument
    """

    def __init__(self, name, required=False, default=NO_DEFAULT):
        # type: (str, bool, object) -> None
        self.name = name
        self.required = required
        if required and default != NO_DEFAULT:
            raise ValueError(
                'If arg is `required=True` then `default` value is redundant, '
                'as caller must always supply value. If you give a `default` '
                'then `required=True` is not needed as arg will always be sent '
                'with a value.'
            )
        self.default = default


class APIEndpoint(object):
    """
    Params from url path (format string), querystring and request headers
    (format string of value portion) will be extracted and made available
    as kwargs on the resulting method call.

    This means arg names must be unique across all three sources of args.
    This is feasible because path and header args can be freely chosen when
    implementing the client (they are just format string identifiers rather
    than part of the REST API itself like querystring args are).

    e.g.

        class ThingServiceClient(APIClient):

            get_things = APIEndpoint(
                'GET',
                '/things/{user_id}/',  # url format string
                querystring_args=(
                    Arg('type', required=True),
                    Arg('offset_id'),
                    Arg('limit', default=25),
                ),
                request_headers={
                    'Authorization': 'Bearer {access_token}'
                }
            )

    This will give you a client with a `get_things` method you can call like:

        response_data = ThingServiceClient.get_things(
            user_id=123,
            type='cat',
            offset_id='65345ff34e344ab53c',
            limit=20,
            access_token='87a64c98b62d39e8625f',
        )

    You can still pass extra args down into the `requests` lib on a per-call
    basis by using `_request_kwargs`:

        response_data = ThingServiceClient.get_things(
            user_id=123,
            type='cat',
            offset_id='65345ff34e344ab53c',
            limit=20,
            access_token='87a64c98b62d39e8625f',
            _request_kwargs={
                'headers': {
                    'X-Depop-WTF': 'something something'
                }
            },
        )

    And for calls with a request body:

        class ThingServiceClient(APIClient):

            new_thing = APIEndpoint(
                'POST',
                '/things/',
                body_required=True,
                body_type=FORM_ENCODED,
                request_headers={
                    'Authorization': 'Bearer {access_token}',
                    'Content-Type': 'application/json; charset=utf-8'
                }
            )

        response_data = ThingServiceClient.new_thing(
            access_token='87a64c98b62d39e8625f',
            body={
                'type': 'dog',
                'name': 'fido',
            }
        )
    """

    RESERVED_NAMES = ('_request_kwargs', '_session')

    method = None  # type: str
    path = None  # type: str

    body_type = None  # type: str
    body_arg = None  # type: str

    request_headers = None  # type: Optional[Dict[str, str]]

    required_args = None  # type: Set[str]
    url_args = None  # type: Set[str]
    querystring_args = None  # type: Set[Arg]
    request_header_args = None  # type: Dict[str, Set[str]]

    def __init__(self,
                 method,  # type: str
                 path,  # type: str
                 querystring_args=None,  # type: Optional[Tuple[Arg, ...]]
                 request_headers=None,  # type: Optional[Dict[str, str]]
                 body_type=BodyType.JSON,  # type: BodyType
                 body_arg='body',  # type: str
                 body_required=False,  # type: bool
                 ):
        # type: (...) -> None
        """
        Kwargs:
            method: 'GET', 'POST' etc
            path: path portion of url (appended to `APIClient.base_url`)
            querystring_args:
                if given, tuple of non-required arg names, or (<arg name>, True)
                for required args
            request_headers: if given, a dict of
                request headers which will be applied to all requests
            body_type: which kwarg in `requests` lib to use for request
                body, e.g. "data" for form-encoded or "json" for JSON encoded
                (in both cases just pass a python dict in the call)
            body_arg: `body_arg` will be made available as a kwarg in the
                resulting client method, as such it needs a unique name. If the
                default name, "body", is needed as a querystring arg then
                specify a custom arg name here.
            body_required: if request must include a body (e.g. POST) then a
                `MissingRequiredArg` error will be raised if called without it.
        """
        f = string.Formatter()

        reserved = set(self.RESERVED_NAMES)
        all_args = set()
        required_args = set()

        # request body arg
        all_args.add(
            _validate_arg_name(body_arg, '`body_arg`', reserved)
        )
        reserved.add(body_arg)
        if body_required:
            required_args.add(body_arg)

        # parse url args
        url_args = set()
        for tokens in f.parse(path):
            if tokens[1] is not None:
                all_args.add(
                    _validate_arg_name(tokens[1], 'Url', reserved)
                )
                url_args.add(tokens[1])
                required_args.add(tokens[1])  # all url args are required

        # parse querystring args
        querystring_args_ = set()
        if querystring_args is not None:
            for arg in querystring_args:
                all_args.add(
                    _validate_arg_name(arg.name, 'Querystring', reserved)
                )
                if arg.required:
                    required_args.add(arg.name)
                querystring_args_.add(arg)

        # parse request-header args
        request_header_args = {}  # type: Dict[str, Set[str]]
        if request_headers is not None:
            for header, value in request_headers.items():
                for tokens in f.parse(value):
                    if tokens[1] is not None:
                        all_args.add(
                            _validate_arg_name(tokens[1], 'Request-header', reserved)
                        )
                        request_header_args.setdefault(header, set()).add(tokens[1])
                        required_args.add(tokens[1])  # all request-header args are required

        self.method = method.lower()
        self.path = path

        self.body_type = body_type.value
        self.body_arg = body_arg

        self.request_headers = request_headers

        self.required_args = required_args
        self.url_args = url_args
        self.querystring_args = querystring_args_
        self.request_header_args = request_header_args


class GetEndpoint(APIEndpoint):

    def __init__(self, *args, **kwargs):
        super(GetEndpoint, self).__init__('GET', *args, **kwargs)


class PostEndpoint(APIEndpoint):

    def __init__(self, *args, **kwargs):
        super(PostEndpoint, self).__init__('POST', *args, **kwargs)


class PatchEndpoint(APIEndpoint):

    def __init__(self, *args, **kwargs):
        super(PatchEndpoint, self).__init__('PATCH', *args, **kwargs)


class PutEndpoint(APIEndpoint):

    def __init__(self, *args, **kwargs):
        super(PutEndpoint, self).__init__('PUT', *args, **kwargs)


class DeleteEndpoint(APIEndpoint):

    def __init__(self, *args, **kwargs):
        super(DeleteEndpoint, self).__init__('DELETE', *args, **kwargs)



import concurrent.futures  # noqa
from functools import partial
from six import add_metaclass
from typing import Any, Dict, Type

from requests_futures.sessions import FuturesSession

from popget.conf import settings
from popget.client import (
    APIClient as BaseAPIClient,
    APIClientMetaclass,
    ConfigClass as BaseConfigClass,
    method_factory,
)
from popget.endpoint import APIEndpoint
from popget.extratypes import ResponseTypes  # noqa
from popget.utils import classproperty


class ConfigClass(BaseConfigClass):

    _async_session = None  # type: FuturesSession

    async_method_template = 'async_{}'

    def __init__(self, config):
        self.async_method_template = getattr(config, 'async_method_template', self.async_method_template)
        super(ConfigClass, self).__init__(config)

    @classproperty
    def async_session(cls):
        # type: () -> FuturesSession
        if not cls._async_session:
            cls._async_session = FuturesSession(session=cls.session)
        return cls._async_session


class AsyncAPIClientMetaclass(APIClientMetaclass):

    config_class = ConfigClass

    @staticmethod
    def add_methods_for_endpoint(methods, name, endpoint, config):
        # type: (Dict[str, classmethod], str, APIEndpoint, Any) -> None
        methods[name] = classmethod(method_factory(endpoint, '_make_request'))
        methods[config.async_method_template.format(name)] = classmethod(
            method_factory(endpoint, '_make_async_request')
        )


@add_metaclass(AsyncAPIClientMetaclass)
class APIClient(BaseAPIClient):

    _config = None  # type: ConfigClass

    @classmethod
    def _make_async_request(cls, method, url, *args, **kwargs):
        # type: (str, str, *Any, **Any) -> concurrent.futures.Future
        """
        Don't call this directly. Instead, add APIEndpoint instances to your
        APIClient sub-class definition. Accessor methods will be generated by
        the APIClientMetaclass magic.

        Kwargs:
            method: name of a method on the `requests` lib, such as 'get'
            url: url of endpoint to call, with all params substituted in
            *args
            **kargs
                passed through to underlying `requests` method

        Returns:
            Future object. Use the standard concurrent.futures API to await
            the result.
        """
        cls._request_kwargs(method, url, args, kwargs)

        call = partial(getattr(cls._config.session, method), url, *args, **kwargs)
        return cls._config.async_session.executor.submit(cls.handle, call, url)



from typing import Any, Dict, List, Union


_T = Any  # should be 'JSONTypes' but mypy doesn't support recursive types yet

JSONTypes = Union[Dict[str, _T], List[_T], str, float, bool, None]

JSONObject = Dict[str, JSONTypes]

JSONList = List[JSONTypes]

ResponseTypes = Union[bytes, JSONTypes]



from functools import partial
from six import add_metaclass
from typing import Any, Callable, Dict, Optional, Set, Tuple, Type, TypeVar, Union  # noqa

from mypy_extensions import Arg, DefaultArg, KwArg
import requests
from requests.exceptions import Timeout

from popget.conf import settings
from popget.endpoint import APIEndpoint, NO_DEFAULT
from popget.errors import MissingRequiredArg
from popget.extratypes import ResponseTypes  # noqa
from popget.utils import get_base_attr, update_nested


ClientMethod = Callable[
    [
        Arg(Type['APIClient'], 'cls'),
        DefaultArg(Optional[Dict[Any, Any]], '_request_kwargs'),
        DefaultArg(Optional[requests.Session], '_session'),
        KwArg(object),
    ],
    Union[ResponseTypes, object]
]


def method_factory(endpoint, client_method_name):
    # type: (APIEndpoint, str) -> ClientMethod
    """
    Kwargs:
        endpoint: the endpoint to generate a callable method for

    Returns:
        A classmethod to be attached to the APIClient, which will perform
        the actual request for this particular endpoint.
        In turn the method returns response content, either string or
        deserialized JSON data.
    """
    def _prepare_request(base_url, _request_kwargs=None, **call_kwargs):
        # type: (str, Optional[Dict], **object) -> Tuple[str, Dict[str, object]]
        """
        Kwargs:
            base_url: base url of API
            _request_kwargs: extra kwargs to pass into the underlying
                `requests` lib method
            **call_kwargs: expected args for the endpoint (url, querystring,
                request-headers etc)
        """
        for required in endpoint.required_args:
            if required not in call_kwargs:
                raise MissingRequiredArg('`{}` arg is required.'.format(required))

        # prepare url kwargs
        url_args = {
            arg: call_kwargs[arg]
            for arg in endpoint.url_args
        }

        # prepare querystring args
        querystring_args = {}
        for arg in endpoint.querystring_args:
            try:
                querystring_args[arg.name] = call_kwargs[arg.name]
            except KeyError:
                # non-required arg
                if arg.default != NO_DEFAULT:
                    querystring_args[arg.name] = arg.default

        # prepare request-header args
        request_headers = {}
        for header, args in endpoint.request_header_args.items():
            header_kwargs = {
                arg: call_kwargs[arg]
                for arg in args
            }
            request_headers[header] = endpoint.request_headers[header].format(**header_kwargs)

        url_template = '{base_url}/{path}'.format(
            base_url=base_url.rstrip('/'),
            path=endpoint.path.lstrip('/'),
        )
        url = url_template.format(**url_args)

        request_kwargs = {
            'params': querystring_args,
            'headers': request_headers,
        }

        request_body = call_kwargs.get(endpoint.body_arg)
        if request_body is not None:
            request_kwargs[endpoint.body_type] = request_body

        if _request_kwargs is not None:
            update_nested(request_kwargs, _request_kwargs)

        return url, request_kwargs

    def client_method(cls,  # type: Type[APIClient]
                      _request_kwargs=None,  # type: Optional[Dict]
                      _session=None,  # type: requests.Session
                      **call_kwargs
                      ):
        # type: (...) -> Union[ResponseTypes, object]
        """
        Returns:
            Response... for non-async clients this will be response content,
            either string or deserialized JSON data if response had
            'application/json' content-type header.
            For async clients this will be some type of future.
        """
        url, request_kwargs = _prepare_request(
            base_url=cls._config.base_url,
            _request_kwargs=_request_kwargs,
            **call_kwargs
        )
        return getattr(cls, client_method_name)(endpoint.method, url, **request_kwargs)

    return client_method


class ConfigClass(object):

    base_url = None  # type: str
    session_cls = None  # type: Type[requests.Session]
    _session = None  # type: requests.Session

    def __init__(self, config):
        self.base_url = getattr(config, 'base_url', '')
        self.session_cls = getattr(config, 'session_cls', requests.Session)

    @property
    def session(self):
        # type: () -> requests.Session
        if not self._session:
            session = self.session_cls()
            session.headers['User-Agent'] = settings.CLIENT_DEFAULT_USER_AGENT
            self._session = session
        return self._session


class APIClientMetaclass(type):
    """
    Makes API call methods from APIEndpoint definitions on the APIClient class

    (this 'metaclass magic' is similar to Django Model class where fields are
    defined on the class, and transformed by the metaclass into usable attrs)

    eg

        class ThingServiceClient(APIClient):

            class Config:
                base_url = 'http://things.depop.com'

            get_things = GetEndpoint(
                '/things/{user_id}/',  # url format string
                (('type', True),),     # required querystring param (validated on call)
            )

    Results in a client method you can call like:

        data = MyAPI.get_things(user_id=2345, type='cat')

    Which will perform a request like:

        GET http://things.depop.com/things/2345/?type=cat

    If response was "Content-Type: application/json" then `data` is already deserialized.

    We use `raise_for_status` so anything >= 400 will raise a `requests.HTTPError`.
    """

    config_class = ConfigClass

    @staticmethod
    def add_methods_for_endpoint(methods, name, endpoint, config):
        # type: (Dict[str, classmethod], str, APIEndpoint, Any) -> None
        methods[name] = classmethod(method_factory(endpoint, '_make_request'))

    def __new__(cls, name, bases, attrs):
        base_config = get_base_attr('Config', bases, attrs)
        attrs['_config'] = config = cls.config_class(base_config)

        methods = {}
        attr_list = list(attrs.items())
        for name, endpoint in attr_list:
            if not isinstance(endpoint, APIEndpoint):
                continue
            cls.add_methods_for_endpoint(methods, name, endpoint, config)
            del attrs[name]
        attrs.update(methods)
        return type.__new__(cls, name, bases, attrs)


def get_response_body(response):
    # type: (requests.Response) -> ResponseTypes
    """
    Kwargs:
        response: from requests lib

    Returns:
        response content - either string or deserialized JSON data if response
        had 'application/json' content type header.
    """

    if response.status_code == 204:
        return None

    try:
        # Requests uses a 'dict-like' case-insensitive dict, but the case-
        # insensitivity only applies to __getitem__ and not .get()
        # So we can't use .get() here but should catch the KeyError.
        content_type = response.headers['Content-Type']
    except KeyError:
        content_type = ''

    if 'application/json' in content_type:
        return response.json()  # JSONTypes
    elif 'text/' in content_type:
        return response.text  # str
    else:
        return response.content  # bytes


@add_metaclass(APIClientMetaclass)
class APIClient(object):

    _config = None  # type: ConfigClass

    @staticmethod
    def _request_kwargs(method, url, args, kwargs):
        # type: (str, str, Tuple[Any, ...], Dict[str, Any]) -> None

        if settings.CLIENT_DISABLE_VERIFY_SSL:
            kwargs['verify'] = False

        # all requests should contain a default timeout value
        if 'timeout' not in kwargs:
            kwargs['timeout'] = settings.CLIENT_TIMEOUT

    @staticmethod
    def handle(call, request_url):
        # type: (Callable[[], requests.Response], str) -> ResponseTypes

        try:
            res = call()
        except Timeout as e:
            # generate a proxy 504 'Gateway Timeout' response
            res = requests.Response()
            res.request = e.request
            res.url = request_url
            res.reason = str(e)
            res.status_code = 504

        res.raise_for_status()
        return get_response_body(res)

    @classmethod
    def _make_request(cls, method, url, session=None, *args, **kwargs):
        # type: (str, str, Optional[requests.Session], *Any, **Any) -> ResponseTypes
        """
        Don't call this directly. Instead, add APIEndpoint instances to your
        APIClient sub-class definition. Accessor methods will be generated by
        the APIClientMetaclass magic.

        Kwargs:
            method: name of a method on the `requests` lib, such as 'get'
            url: url of endpoint to call, with all params substituted in
            *args
            **kargs
                passed through to underlying `requests` method

        Returns:
            response content, either string or deserialized JSON data if
            response had 'application/json' content type header.
        """
        cls._request_kwargs(method, url, args, kwargs)
        if session is None:
            session = cls._config.session
        call = partial(getattr(session, method), url, *args, **kwargs)
        return cls.handle(call, url)



import collections
from typing import AnyStr, Dict, Optional, Tuple, Type


def update_nested(d, u):
    # type: (Dict, collections.Mapping) -> Dict
    """
    https://stackoverflow.com/a/3233356/202168
    A dict update that supports nested keys without overwriting the whole
    sub-dict like the built-in `update` method of dicts.

    e.g.

    mydict = {
        'a': {
            'b': 1
        }
    }
    mydict.update({'a': {'c': 2}})

    assert mydict == {
        'a': {
            'b': 1,
            'c': 2,
        }
    }
    """
    for k, v in u.items():
        if isinstance(v, collections.Mapping):
            r = update_nested(d.get(k, {}), v)
            d[k] = r
        else:
            d[k] = u[k]
    return d


def get_base_attr(attr, bases, attrs):
    # type: (str, Tuple[Type, ...], Dict[AnyStr, object]) -> Optional[object]
    """
    Given an attr name, recursively look through the given base classes
    until finding one where that attr is present, returning the value.
    """
    try:
        return attrs[attr]
    except KeyError:
        pass
    for base in bases:
        try:
            return get_base_attr(attr, base.__bases__, base.__dict__)
        except KeyError:
            pass
    return None


class classproperty(property):
    def __get__(self, cls, owner):
        return classmethod(self.fget).__get__(None, owner)()



import os
from typing import Optional

from popget.__about__ import __version__


# namespace for config keys loaded from e.g. Django conf or env vars
CONFIG_NAMESPACE = os.getenv('POPGET_CONFIG_NAMESPACE', 'POPGET')

# optional import path to file containing namespaced config (e.g. 'django.conf.settings')
APP_CONFIG = os.getenv('POPGET_APP_CONFIG', None)


CLIENT_DEFAULT_USER_AGENT = 'popget/{}'.format(__version__)

CLIENT_TIMEOUT = None  # type: Optional[float]

CLIENT_DISABLE_VERIFY_SSL = False



from setuptools import setup
from codecs import open  # To use a consistent encoding
from os import path


here = path.abspath(path.dirname(__file__))


# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

about = {}
with open(path.join(here, 'popget', '__about__.py'), 'r', 'utf-8') as f:
    exec(f.read(), about)


setup(
    name='popget',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=about['__version__'],

    description='Simple REST-API client for Python.',
    long_description=long_description,

    url='https://github.com/depop/popget',

    author='Depop',
    author_email='dev@depop.com',

    license='Apache 2.0',
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 2.7',
    ],
    install_requires=[
        'flexisettings>=1.0,<1.1',
        'requests>2.0,<3.0',
        'typing>=3.6.2,<4.0',
        'six',
        'enum34>=1.1.6,<2.0.0; python_version < "3.4"',
        'futures>=3.1.1,<3.2.0; python_version < "3.2"',
        'mypy_extensions==0.3.0; python_version < "3.2"',
    ],
    extras_require={
        'threadpool':  ['requests-futures>=0.9.7,<1.0.0'],
    },

    packages=[
        'popget',
        'popget.async',
        'popget.conf',
    ],
)
